@@ -1,7 +1,7 @@
 ﻿/*
- *  This file is part of Poedit (http://poedit.net)
+ *  This file is part of Poedit (https://poedit.net)
  *
- *  Copyright (C) 2000-2015 Vaclav Slavik
+ *  Copyright (C) 2000-2016 Vaclav Slavik
  *
  *  Permission is hereby granted, free of charge, to any person obtaining a
  *  copy of this software and associated documentation files (the "Software"),
@@ -33,6 +33,7 @@
 #include <wx/checkbox.h>
 #include <wx/choice.h>
 #include <wx/checklst.h>
+#include <wx/notebook.h>
 #include <wx/sizer.h>
 #include <wx/stattext.h>
 #include <wx/fontutil.h>
@@ -48,10 +49,15 @@
 #include <wx/xrc/xmlres.h>
 #include <wx/numformatter.h>
 
+#if !wxCHECK_VERSION(3,1,0)
+    #define CenterVertical() Center()
+#endif
+
 #include "prefsdlg.h"
 #include "edapp.h"
 #include "edframe.h"
 #include "catalog.h"
+#include "crowdin_gui.h"
 #include "hidpi.h"
 #include "tm/transmem.h"
 #include "chooselang.h"
@@ -60,6 +66,7 @@
 #include "spellchecking.h"
 #include "utility.h"
 #include "customcontrols.h"
+#include "unicode_helpers.h"
 
 #ifdef __WXMSW__
 #include <winsparkle.h>
@@ -80,7 +87,7 @@ class PrefsPanel : public wxPanel
 {
 public:
     PrefsPanel(wxWindow *parent)
-        : wxPanel(parent), m_supressDataTransfer(0)
+        : wxPanel(parent), m_suppressDataTransfer(0)
     {
 #ifdef __WXOSX__
         // Refresh the content of prefs panels when re-opening it.
@@ -100,14 +107,14 @@ public:
 
     bool TransferDataToWindow() override
     {
-        if (m_supressDataTransfer)
+        if (m_suppressDataTransfer)
             return false;
-        m_supressDataTransfer++;
+        m_suppressDataTransfer++;
         InitValues(*wxConfig::Get());
-        m_supressDataTransfer--;
+        m_suppressDataTransfer--;
 
         // This is a "bit" of a hack: we take advantage of being in the last point before
-        // showing the window and re-layout it on the off chance that some data transfered
+        // showing the window and re-layout it on the off chance that some data transferred
         // into the window affected its size. And, currently more importantly, to reflect
         // ExplanationLabel instances' rewrapping.
         Fit();
@@ -119,11 +126,11 @@ public:
 
     bool TransferDataFromWindow() override
     {
-        if (m_supressDataTransfer)
+        if (m_suppressDataTransfer)
             return false;
-        m_supressDataTransfer++;
+        m_suppressDataTransfer++;
         SaveValues(*wxConfig::Get());
-        m_supressDataTransfer--;
+        m_suppressDataTransfer--;
         return true;
     }
 
@@ -137,18 +144,10 @@ protected:
     virtual void InitValues(const wxConfigBase& cfg) = 0;
     virtual void SaveValues(wxConfigBase& cfg) = 0;
 
-    int m_supressDataTransfer;
+    int m_suppressDataTransfer;
 };
 
 
-#ifdef __WXOSX__
-    #define BORDER_WIN(dir, n) Border(dir, 0)
-    #define BORDER_OSX(dir, n) Border(dir, n)
-#else
-    #define BORDER_WIN(dir, n) Border(dir, n)
-    #define BORDER_OSX(dir, n) Border(dir, 0)
-#endif
-
 
 class GeneralPageWindow : public PrefsPanel
 {
@@ -170,15 +169,15 @@ public:
         sizer->Add(translator, wxSizerFlags().Expand());
 
         auto nameLabel = new wxStaticText(this, wxID_ANY, _("Name:"));
-        translator->Add(nameLabel, wxSizerFlags().Center().Right().BORDER_OSX(wxTOP, 1));
+        translator->Add(nameLabel, wxSizerFlags().CenterVertical().Right().BORDER_OSX(wxTOP, 1));
         m_userName = new wxTextCtrl(this, wxID_ANY);
         m_userName->SetHint(_("Your Name"));
-        translator->Add(m_userName, wxSizerFlags(1).Expand().Center());
+        translator->Add(m_userName, wxSizerFlags(1).Expand().CenterVertical());
         auto emailLabel = new wxStaticText(this, wxID_ANY, _("Email:"));
-        translator->Add(emailLabel, wxSizerFlags().Center().Right().BORDER_OSX(wxTOP, 1));
+        translator->Add(emailLabel, wxSizerFlags().CenterVertical().Right().BORDER_OSX(wxTOP, 1));
         m_userEmail = new wxTextCtrl(this, wxID_ANY);
         m_userEmail->SetHint(_("your_email@example.com"));
-        translator->Add(m_userEmail, wxSizerFlags(1).Expand().Center());
+        translator->Add(m_userEmail, wxSizerFlags(1).Expand().CenterVertical());
         translator->AddSpacer(PX(1));
         translator->Add(new ExplanationLabel(this, _("Your name and email address are only used to set the Last-Translator header of GNU gettext files.")), wxSizerFlags(1).Expand().PXBorder(wxRIGHT));
 #ifdef __WXOSX__
@@ -225,10 +224,10 @@ public:
         m_fontText = new wxFontPickerCtrl(this, wxID_ANY);
         m_fontText->SetMinSize(wxSize(PX(120), -1));
 
-        appearance->Add(m_useFontList, wxSizerFlags().Center().Left());
-        appearance->Add(m_fontList, wxSizerFlags().Center().Expand());
-        appearance->Add(m_useFontText, wxSizerFlags().Center().Left());
-        appearance->Add(m_fontText, wxSizerFlags().Center().Expand());
+        appearance->Add(m_useFontList, wxSizerFlags().CenterVertical().Left());
+        appearance->Add(m_fontList, wxSizerFlags().CenterVertical().Expand());
+        appearance->Add(m_useFontText, wxSizerFlags().CenterVertical().Left());
+        appearance->Add(m_fontText, wxSizerFlags().CenterVertical().Expand());
 
 #if NEED_CHOOSELANG_UI 
         m_uiLanguage = new wxButton(this, wxID_ANY, _("Change UI language"));
@@ -286,7 +285,7 @@ public:
         m_useFontText->SetValue(cfg.ReadBool("custom_font_text_use", false));
 
         #if defined(__WXOSX__)
-            #define DEFAULT_FONT "Helvetica"
+            #define DEFAULT_FONT "Helvetica Neue"
         #elif defined(__WXMSW__)
             #define DEFAULT_FONT "Arial"
         #elif defined(__WXGTK__)
@@ -398,7 +397,7 @@ public:
         auto explain = new ExplanationLabel(this, explainTxt);
         sizer->Add(explain, wxSizerFlags().Expand().Border(wxLEFT, PX(ExplanationLabel::CHECKBOX_INDENT)));
 
-        auto learnMore = new LearnMoreLink(this, "http://poedit.net/trac/wiki/Doc/TranslationMemory");
+        auto learnMore = new LearnMoreLink(this, "https://poedit.net/trac/wiki/Doc/TranslationMemory");
         sizer->AddSpacer(PX(5));
         sizer->Add(learnMore, wxSizerFlags().Border(wxLEFT, PX(ExplanationLabel::CHECKBOX_INDENT + LearnMoreLink::EXTRA_INDENT)));
         sizer->AddSpacer(PX(10));
@@ -473,8 +472,7 @@ private:
             _("Select translation files to import"),
             wxEmptyString,
             wxEmptyString,
-			wxString::Format("%s (*.po)|*.po|%s (*.*)|*.*",
-                _("PO Translation Files"), _("All Files")),
+			Catalog::GetTypesFileMask({Catalog::Type::PO}),
 			wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_MULTIPLE));
 
         dlg->ShowWindowModalThenDo([=](int retcode){
@@ -489,15 +487,15 @@ private:
                                       (int)paths.size() * 2 + 1,
                                       this,
                                       wxPD_APP_MODAL|wxPD_AUTO_HIDE|wxPD_CAN_ABORT);
-            auto tm = TranslationMemory::Get().CreateWriter();
+            auto tm = TranslationMemory::Get().GetWriter();
             int step = 0;
             for (size_t i = 0; i < paths.size(); i++)
             {
-                std::unique_ptr<Catalog> cat(new Catalog(paths[i]));
+                CatalogPtr cat = std::make_shared<Catalog>(paths[i]);
                 if (!progress.Update(++step))
                     break;
                 if (cat->IsOk())
-                    tm->Insert(*cat);
+                    tm->Insert(cat);
                 if (!progress.Update(++step))
                     break;
             }
@@ -520,7 +518,7 @@ private:
         dlg->ShowWindowModalThenDo([this,dlg](int retcode){
             if (retcode == wxID_YES) {
                 wxBusyCursor bcur;
-                auto tm = TranslationMemory::Get().CreateWriter();
+                auto tm = TranslationMemory::Get().GetWriter();
                 tm->DeleteAll();
                 tm->Commit();
                 UpdateStats();
@@ -542,7 +540,7 @@ class TMPage : public wxPreferencesPage
 public:
     wxString GetName() const override
     {
-#ifdef __WXOSX__
+#if defined(__WXOSX__) || defined(__WXGTK__)
         // TRANSLATORS: This is abbreviation of "Translation Memory" used in Preferences on OS X.
         // Long text looks weird there, too short (like TM) too, but less so. "General" is about ideal
         // length there.
@@ -576,7 +574,7 @@ public:
         sizer->Add(horizontal, wxSizerFlags(1).Expand());
 
         m_list = new wxCheckListBox(this, wxID_ANY);
-        m_list->SetMinSize(wxSize(PX(250),PX(300)));
+        m_list->SetMinSize(wxSize(PX(300),PX(300)));
 #ifdef __WXOSX__
         m_list->SetWindowVariant(wxWINDOW_VARIANT_SMALL);
 #endif
@@ -609,7 +607,7 @@ public:
         m_list->Clear();
         for (const auto& item: m_extractors.Data)
         {
-            auto index = m_list->Append(item.Name);
+            auto index = m_list->Append(bidi::platform_mark_direction(item.Name));
             m_list->Check(index, item.Enabled);
         }
 
@@ -626,7 +624,7 @@ public:
     }
 
 private:
-    /// Called to launch dialog for editting parser properties.
+    /// Called to launch dialog for editing parser properties.
     template<typename TFunctor>
     void EditExtractor(int num, TFunctor completionHandler)
     {
@@ -642,12 +640,12 @@ private:
 
         {
             const Extractor& nfo = m_extractors.Data[num];
-            extractor_language->SetValue(nfo.Name);
-            extractor_extensions->SetValue(nfo.Extensions);
-            extractor_command->SetValue(nfo.Command);
-            extractor_keywords->SetValue(nfo.KeywordItem);
-            extractor_files->SetValue(nfo.FileItem);
-            extractor_charset->SetValue(nfo.CharsetItem);
+            extractor_language->SetValue(bidi::platform_mark_direction(nfo.Name));
+            extractor_extensions->SetValue(bidi::mark_direction(nfo.Extensions, TextDirection::LTR));
+            extractor_command->SetValue(bidi::mark_direction(nfo.Command, TextDirection::LTR));
+            extractor_keywords->SetValue(bidi::mark_direction(nfo.KeywordItem, TextDirection::LTR));
+            extractor_files->SetValue(bidi::mark_direction(nfo.FileItem, TextDirection::LTR));
+            extractor_charset->SetValue(bidi::mark_direction(nfo.CharsetItem, TextDirection::LTR));
         }
 
         dlg->Bind
@@ -663,19 +661,19 @@ private:
             wxID_OK
         );
 
-        m_supressDataTransfer++;
+        m_suppressDataTransfer++;
         dlg->ShowWindowModalThenDo([=](int retcode){
-            m_supressDataTransfer--;
+            m_suppressDataTransfer--;
             (void)dlg; // force use
             if (retcode == wxID_OK)
             {
                 Extractor& nfo = m_extractors.Data[num];
-                nfo.Name = extractor_language->GetValue().Strip(wxString::both);
-                nfo.Extensions = extractor_extensions->GetValue().Strip(wxString::both);
-                nfo.Command = extractor_command->GetValue().Strip(wxString::both);
-                nfo.KeywordItem = extractor_keywords->GetValue().Strip(wxString::both);
-                nfo.FileItem = extractor_files->GetValue().Strip(wxString::both);
-                nfo.CharsetItem = extractor_charset->GetValue().Strip(wxString::both);
+                nfo.Name = bidi::strip_control_chars(extractor_language->GetValue().Strip(wxString::both));
+                nfo.Extensions = bidi::strip_control_chars(extractor_extensions->GetValue().Strip(wxString::both));
+                nfo.Command = bidi::strip_control_chars(extractor_command->GetValue().Strip(wxString::both));
+                nfo.KeywordItem = bidi::strip_control_chars(extractor_keywords->GetValue().Strip(wxString::both));
+                nfo.FileItem = bidi::strip_control_chars(extractor_files->GetValue().Strip(wxString::both));
+                nfo.CharsetItem = bidi::strip_control_chars(extractor_charset->GetValue().Strip(wxString::both));
                 m_list->SetString(num, nfo.Name);
             }
             completionHandler(retcode == wxID_OK);
@@ -684,7 +682,7 @@ private:
 
     void OnNewExtractor(wxCommandEvent&)
     {
-        m_supressDataTransfer++;
+        m_suppressDataTransfer++;
 
         Extractor info;
         m_extractors.Data.push_back(info);
@@ -702,7 +700,7 @@ private:
                 m_extractors.Data.erase(m_extractors.Data.begin() + index);
             }
 
-            m_supressDataTransfer--;
+            m_suppressDataTransfer--;
 
             if (wxPreferencesEditor::ShouldApplyChangesImmediately())
                 TransferDataFromWindow();
@@ -720,11 +718,26 @@ private:
     void OnDeleteExtractor(wxCommandEvent&)
     {
         int index = m_list->GetSelection();
-        m_extractors.Data.erase(m_extractors.Data.begin() + index);
-        m_list->Delete(index);
 
-        if (wxPreferencesEditor::ShouldApplyChangesImmediately())
-            TransferDataFromWindow();
+        auto title = _("Delete extractor");
+        auto main = wxString::Format(_(L"Are you sure you want to delete the “%s” extractor?"), m_extractors.Data[index].Name);
+
+        wxWindowPtr<wxMessageDialog> dlg(new wxMessageDialog(this, main, title, wxYES_NO | wxNO_DEFAULT | wxICON_QUESTION));
+        #ifdef __WXOSX__
+        dlg->SetExtendedMessage(" "); // prevent wx from using the title stupidly
+        #endif
+        dlg->SetYesNoLabels(_("Delete"), _("Cancel"));
+
+        dlg->ShowWindowModalThenDo([this,index,dlg](int retcode){
+            if (retcode == wxID_YES)
+            {
+                m_extractors.Data.erase(m_extractors.Data.begin() + index);
+                m_list->Delete(index);
+
+                if (wxPreferencesEditor::ShouldApplyChangesImmediately())
+                    TransferDataFromWindow();
+            }
+        });
     }
 
     void OnEnableExtractor(wxCommandEvent& e)
@@ -752,6 +765,58 @@ public:
 
 
 
+#ifdef HAVE_HTTP_CLIENT
+class AccountsPageWindow : public PrefsPanel
+{
+public:
+    AccountsPageWindow(wxWindow *parent) : PrefsPanel(parent)
+    {
+        wxSizer *topsizer = new wxBoxSizer(wxVERTICAL);
+        m_login = new CrowdinLoginPanel(this);
+        topsizer->Add(m_login, wxSizerFlags(1).Expand().Border(wxALL, PX(15)));
+        SetSizer(topsizer);
+
+    #ifdef __WXOSX__
+        // This window was created on demand, init immediately:
+        m_login->EnsureInitialized();
+    #else
+        // On other platforms, notebook pages are all created at once. Don't do
+        // the expensive initialization until shown for the first time. This code
+        // is a hack that takes advantage of wxPreferencesEditor's implementation
+        // detail, but oh well:
+        auto notebook = dynamic_cast<wxNotebook*>(parent);
+        if (notebook)
+        {
+            notebook->Bind(wxEVT_NOTEBOOK_PAGE_CHANGED, [=](wxBookCtrlEvent& e){
+                e.Skip();
+                if (notebook->GetPage(e.GetSelection()) == this)
+                    m_login->EnsureInitialized();
+            });
+        }
+    #endif
+    }
+
+    void InitValues(const wxConfigBase&) override
+    {
+    }
+
+    void SaveValues(wxConfigBase&) override
+    {
+    }
+
+private:
+    CrowdinLoginPanel *m_login;
+};
+
+class AccountsPage : public wxPreferencesPage
+{
+public:
+    wxString GetName() const override { return PXNotebookTab(_("Accounts")); }
+    wxBitmap GetLargeIcon() const override { return wxArtProvider::GetBitmap("Prefs-Accounts"); }
+    wxWindow *CreateWindow(wxWindow *parent) override { return new AccountsPageWindow(parent); }
+};
+#endif // HAVE_HTTP_CLIENT
+
 
 #ifdef HAS_UPDATES_CHECK
 class UpdatesPageWindow : public PrefsPanel
@@ -760,7 +825,7 @@ public:
     UpdatesPageWindow(wxWindow *parent) : PrefsPanel(parent)
     {
         wxSizer *topsizer = new wxBoxSizer(wxVERTICAL);
-        topsizer->SetMinSize(PX(350), -1); // for OS X look, wouldn't fit the toolbar otherwise
+        topsizer->SetMinSize(PX(400), -1); // for OS X look, wouldn't fit the toolbar otherwise
 
         wxSizer *sizer = new wxBoxSizer(wxVERTICAL);
         topsizer->Add(sizer, wxSizerFlags().Expand().PXDoubleBorderAll());
@@ -847,7 +912,11 @@ public:
         m_wrap = new wxCheckBox(this, wxID_ANY, _("Wrap at:"));
         crlfbox->AddSpacer(PX(10));
         crlfbox->Add(m_wrap, wxSizerFlags().Center().BORDER_WIN(wxTOP, PX(1)));
+    #ifdef __WXGTK3__
+        m_wrapWidth = new wxSpinCtrl(this, wxID_ANY, "", wxDefaultPosition, wxSize(PX(100),-1));
+    #else
         m_wrapWidth = new wxSpinCtrl(this, wxID_ANY, "", wxDefaultPosition, wxSize(PX(50),-1));
+    #endif
         m_wrapWidth->SetRange(10, 1000);
         crlfbox->Add(m_wrapWidth, wxSizerFlags().Center().BORDER_OSX(wxLEFT, PX(3)));
 
@@ -914,10 +983,18 @@ public:
 
 std::unique_ptr<PoeditPreferencesEditor> PoeditPreferencesEditor::Create()
 {
+#if 0
+    // TRANSLATORS: Title of the preferences window on Windows and Linux. "%s" is replaced with "Poedit" when running.
+    _("%s Preferences");
+#endif
+
     std::unique_ptr<PoeditPreferencesEditor> p(new PoeditPreferencesEditor);
     p->AddPage(new GeneralPage);
     p->AddPage(new TMPage);
     p->AddPage(new ExtractorsPage);
+#ifdef HAVE_HTTP_CLIENT
+    p->AddPage(new AccountsPage);
+#endif
 #ifdef HAS_UPDATES_CHECK
     p->AddPage(new UpdatesPage);
 #endif
